/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.dyn;

import java.lang.reflect.Field;
import java.lang.reflect.Method;

import sun.misc.Unsafe;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;

public class CoroutineSupport {
	private static final int COROUTINE_STACK_HASH_MULT = 37;

	// Controls debugging and tracing, for maximum performance the actual if(DEBUG/TRACE) code needs to be commented out
	static final boolean DEBUG = false;
	static final boolean TRACE = false;

	static final Object TERMINATED = new Object();

	// All the stacks that belong to this thread, a 0 entry means that the slot is unoccupied
	private long[] stacks;

	// The thread that this CoroutineSupport belongs to. There's only one CoroutineSupport per Thread
	private final Thread thread;
	// The initial coroutine of the Thread
	private final Coroutine threadCoroutine;

	// The currently executing, symmetric or asymmetric coroutine
	CoroutineBase currentCoroutine;
	// The anchor of the doubly-linked ring of coroutines
	Coroutine scheduledCoroutines;

	// The last running coroutine (which needs to be released by setting its context)
	CoroutineBase oldCoroutine;

	private static final long coroContextOffset;
	private static final Unsafe unsafe;

	final Object RUNNING = new Object();
	final Object LOCKED = new Object();

	static {
		registerNatives();
		coroContextOffset = getFieldOffset(CoroutineBase.class, "context");
		unsafe = Unsafe.getUnsafe();
	}
	
	private static int getIntConfig(String key, int defaultValue) {
		String value = System.getProperty(key);
		if (value != null) {
			try {
				return Integer.valueOf(value);
			} catch (NumberFormatException e) {
				// nothing to do
			}
		}
		return defaultValue;
	}

	private static long getFieldOffset(Class<?> clazz, String name) {
		try {
			Field f = clazz.getDeclaredField(name);
			return Unsafe.getUnsafe().objectFieldOffset(f);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public CoroutineSupport(Thread thread) {
		if (thread.getCoroutineSupport() != null)
			throw new IllegalArgumentException("Cannot instantiate CoroutineThreadSupport for existing Thread");
		this.stacks = new long[getIntConfig("coro.stacksPerThread", 100)];
		this.thread = thread;
		threadCoroutine = new Coroutine(this, getThreadCoroutine());
		threadCoroutine.next = threadCoroutine;
		threadCoroutine.last = threadCoroutine;
		currentCoroutine = threadCoroutine;
		scheduledCoroutines = threadCoroutine;
	}

	private int stackHash(CoroutineBase coro) {
		// This makes sure that hash isn't negative because Math.abs isn't enough!
		int hash = (coro.hashCode() * COROUTINE_STACK_HASH_MULT) >>> 1;
		return hash % stacks.length;
	}

	private long getOrCreateStack(CoroutineBase coro, long stacksize) {
		int index = stackHash(coro);
		long stack = stacks[index];
		if (stack == 0) {
			stack = createCoroutineStack(stacksize);
			if (DEBUG)
				System.out.println("Created new stack " + stack + " for index " + index);
			stacks[index] = stack;
		}
		return stack;
	}

	void addCoroutine(Coroutine coroutine, long stacksize) {
		assert scheduledCoroutines != null;
		assert currentCoroutine != null;

		long stack = getOrCreateStack(coroutine, stacksize);
		coroutine.data = createCoroutine(coroutine, stack);
		if (DEBUG)
			System.out.println("add Coroutine " + coroutine + " using stack: " + stack + ", data" + coroutine.data);

		// add the coroutine into the doubly linked ring
		coroutine.next = scheduledCoroutines.next;
		coroutine.last = scheduledCoroutines;
		scheduledCoroutines.next = coroutine;
		coroutine.next.last = coroutine;
	}

	void addCoroutine(AsymCoroutine<?, ?> coroutine, long stacksize) {
		long stack = getOrCreateStack(coroutine, stacksize);
		coroutine.data = createCoroutine(coroutine, stack);
		if (DEBUG)
			System.out.println("add AsymCoroutine " + coroutine + " using stack: " + stack + ", data" + coroutine.data);

		coroutine.caller = null;
	}

	Thread getThread() {
		return thread;
	}

	private boolean isDisposable(AsymCoroutine<?, ?> coroutine) {
		// TODO implement
		return true;
	}

	public void drain() {
		if (Thread.currentThread() != thread)
			throw new IllegalArgumentException("Cannot drain another threads CoroutineThreadSupport");

		if (DEBUG)
			System.out.println("draining");
		try {
			// drain all scheduled coroutines
			while (scheduledCoroutines.next != scheduledCoroutines)
				symmetricExitInternal(scheduledCoroutines.next);

			CoroutineBase coro;
			while ((coro = cleanupCoroutine()) != null) {
				System.out.println(coro);
				throw new NotImplementedException();
			}
		} catch (Throwable t) {
			t.printStackTrace();
		}

		// free all stack associated with this thread
		for (long stack : stacks) {
			if (stack != 0) {
				freeCoroutineStack(stack);
			}
		}
		stacks = null;
	}

	final void symmetricStart(Coroutine coroutine) {
		oldCoroutine.context = this;
	}

	final void asymmetricStart(AsymCoroutine<?, ?> coroutine) {
	}

	void symmetricYield() {
		if (scheduledCoroutines != currentCoroutine)
			throw new IllegalThreadStateException("Cannot call yield from within an asymmetric coroutine");
		assert currentCoroutine instanceof Coroutine;

		if (TRACE)
			System.out.println("locking for symmetric yield...");

		Coroutine next;
		do {
			next = scheduledCoroutines.next;
			if (next == scheduledCoroutines)
				return;
		} while (!unsafe.compareAndSwapObject(next, coroContextOffset, this, RUNNING));

		if (TRACE)
			System.out.println("symmetric yield to " + next);

		prepareSwitch(next);
		final Coroutine current = scheduledCoroutines;
		oldCoroutine = current;
		scheduledCoroutines = next;
		currentCoroutine = next;

		switchTo(oldCoroutine, next);

		finishCoroutineSwitch(current.threadSupport);
	}

	public void symmetricYieldTo(Coroutine target) {
		if (!unsafe.compareAndSwapObject(target, coroContextOffset, this, RUNNING)) {
			coroutineLockFailed(target);
			return;
		}
		if (scheduledCoroutines != currentCoroutine) {
			target.context = this;
			throw new IllegalThreadStateException("Cannot call yield from within an asymmetric coroutine");
		}
		assert currentCoroutine instanceof Coroutine;

		moveCoroutine(scheduledCoroutines, target);

		prepareSwitch(target);
		final Coroutine current = scheduledCoroutines;
		oldCoroutine = current;
		scheduledCoroutines = target;
		currentCoroutine = target;

		switchTo(oldCoroutine, target);

		finishCoroutineSwitch(current.threadSupport);
	}

	private void coroutineLockFailed(CoroutineBase target) {
		if (target.context == RUNNING) {
			assert target == scheduledCoroutines;
		} else if (target.context == TERMINATED)
			throw new IllegalArgumentException("cannot yield to terminated coroutine");
		else if (target.context instanceof CoroutineSupport)
			throw new IllegalArgumentException("cannot yield to a coroutine that belongs to another thread");
		else {
			throw new IllegalArgumentException("cannot yield to coroutine, context = " + target.context);
		}
	}

	private void moveCoroutine(Coroutine a, Coroutine position) {
		assert a.context != this && position.context != this;

		// remove a from the ring
		a.last.next = a.next;
		a.next.last = a.last;

		// ... and insert at the new position
		a.next = position.next;
		a.last = position;
		a.next.last = a;
		position.next = a;
	}

	public void symmetricStopCoroutine(Coroutine target) {
		if (!unsafe.compareAndSwapObject(target, coroContextOffset, this, RUNNING)) {
			coroutineLockFailed(target);
			return;
		}
		if (scheduledCoroutines != currentCoroutine) {
			target.context = this;
			throw new IllegalThreadStateException("Cannot call yield from within an asymmetric coroutine");
		}
		assert currentCoroutine instanceof Coroutine;

		moveCoroutine(scheduledCoroutines, target);

		prepareSwitch(target);
		final Coroutine current = scheduledCoroutines;
		oldCoroutine = current;
		scheduledCoroutines = target;
		currentCoroutine = target;

		switchToAndExit(oldCoroutine, target);

		finishCoroutineSwitch(current.threadSupport);
	}

	void symmetricExitInternal(Coroutine coroutine) {
		if (scheduledCoroutines != currentCoroutine)
			throw new IllegalThreadStateException("Cannot call exitNext from within an unscheduled coroutine");
		assert currentCoroutine instanceof Coroutine;
		assert currentCoroutine != coroutine;

		// remove the coroutine from the ring
		coroutine.last.next = coroutine.next;
		coroutine.next.last = coroutine.last;

		if (!isDisposable(coroutine.data)) {
			// and insert it before the current coroutine
			coroutine.last = scheduledCoroutines.last;
			coroutine.next = scheduledCoroutines;
			coroutine.last.next = coroutine;
			scheduledCoroutines.last = coroutine;

			prepareSwitch(coroutine);
			oldCoroutine = scheduledCoroutines;
			scheduledCoroutines = coroutine;
			currentCoroutine = coroutine;
			switchToAndExit(oldCoroutine, coroutine);
		}
	}

	void asymmetricCall(AsymCoroutine<?, ?> target) {
		if (target.threadSupport != this) {
			tryMigrateCoroutine(target);
		}

		if (!unsafe.compareAndSwapObject(target, coroContextOffset, this, RUNNING)) {
			coroutineLockFailed(target);
			return;
		}

		if (target.caller != null)
			throw new IllegalArgumentException("Coroutine already in use");
		if (target.data == 0)
			throw new IllegalArgumentException("Target coroutine has already finished");
		// if (TRACE)
		// System.out.println("yieldCall " + target + " (" + target.data + ")");

		prepareSwitch(target);
		final CoroutineBase current = currentCoroutine;
		target.caller = current;
		currentCoroutine = target;
		switchTo(target.caller, target);

		finishCoroutineSwitch(current.threadSupport);
	}

	private final static void dummy() {
	}

	private static final void finishCoroutineSwitch(CoroutineSupport support) {
		dummy();
		support.oldCoroutine.context = support;
		support.oldCoroutine = null;
	}

	private final void finishCoroutineSwitch() {
		dummy();
		oldCoroutine.context = this;
		oldCoroutine = null;
	}

	private final void finishCoroutineSwitch2(Object dummy) {
		dummy();
		oldCoroutine.context = this;
		oldCoroutine = null;
	}

	void asymmetricReturn(final AsymCoroutine<?, ?> current) {
		if (current != currentCoroutine)
			throw new IllegalThreadStateException("cannot return from non-current fiber");
		final CoroutineBase caller = current.caller;
		// if (TRACE)
		// System.out.println("yieldReturn " + caller + " (" + caller.data + ")");

		prepareSwitch(caller);
		current.caller = null;
		oldCoroutine = current;
		currentCoroutine = caller;
		switchTo(current, currentCoroutine);
	}

	void asymmetricReturnAndTerminate(final AsymCoroutine<?, ?> current) {
		if (current != currentCoroutine)
			throw new IllegalThreadStateException("cannot return from non-current fiber");
		final CoroutineBase caller = current.caller;
		if (TRACE)
			System.out.println("yieldReturn " + caller + " (" + caller.data + ")");

		prepareSwitch(caller);
		oldCoroutine = current;
		current.caller = null;
		currentCoroutine = caller;
		switchToAndTerminate(current, currentCoroutine);
	}

	private void tryMigrateCoroutine(AsymCoroutine<?, ?> target) {
		throw new IllegalArgumentException("Cannot activate a coroutine that belongs to another thread");
	}

	void terminateCoroutine() {
		assert currentCoroutine == scheduledCoroutines;
		assert currentCoroutine != threadCoroutine : "cannot exit thread coroutine";
		assert scheduledCoroutines != scheduledCoroutines.next : "last coroutine shouldn't call coroutineexit";

		Coroutine old = scheduledCoroutines;
		oldCoroutine = old;
		Coroutine forward = old.next;
		prepareSwitch(forward);
		currentCoroutine = forward;
		scheduledCoroutines = forward;
		old.last.next = old.next;
		old.next.last = old.last;

		if (DEBUG)
			System.out.println("to be terminated: " + old);
		switchToAndTerminate(old, forward);
	}

	void terminateCallable() {
		assert currentCoroutine != scheduledCoroutines;
		assert currentCoroutine instanceof AsymCoroutine<?, ?>;

		assert currentCoroutine.context == RUNNING;
		oldCoroutine = currentCoroutine;

		if (DEBUG)
			System.out.println("to be terminated: " + currentCoroutine);
		asymmetricReturnAndTerminate((AsymCoroutine<?, ?>) currentCoroutine);
	}

	public boolean stealCoroutine(CoroutineBase coroutine) {
		CoroutineSupport other = coroutine.threadSupport;
		if (coroutine == other.threadCoroutine) {
			throw new IllegalArgumentException("Cannot steal a threads initial coroutine");
		}
		if (other == this) {
			throw new IllegalArgumentException("Cannot steal coroutine from current thread");
		}

		if (coroutine.isFinished()) {
			throw new IllegalArgumentException("Cannot serialize a coroutine that is finished");
		}
		
		Coroutine last = null;
		if (coroutine instanceof Coroutine) {
			Coroutine coro = (Coroutine) coroutine;
			last = coro.last;

			if (!unsafe.compareAndSwapObject(last, coroContextOffset, other, null)) {
				if (TRACE)
					System.out.println("unable to lock coroutine for stealing");
				return false;
			}
		}

		if (!unsafe.compareAndSwapObject(coroutine, coroContextOffset, other, null)) {
			last.context = other;
			if (TRACE)
				System.out.println("unable to lock coroutine for stealing");
			return false;
		}
		if (coroutine instanceof Coroutine && ((Coroutine) coroutine).last != last) {
			last.context = other;
			coroutine.context = other;
			if (TRACE)
				System.out.println("unable to lock coroutine for stealing");
		}

		long stack = getOrCreateStack(coroutine, -1);
		if (!stealCoroutineData(coroutine.data, stack, coroutine.threadSupport.thread, this)) {
			// something went wrong while stealing the native data structures - restore the coroutine
			if (TRACE)
				System.out.println("unable to steal coroutine native data structures");
			coroutine.context = other;
			return false;
		}
		if (TRACE)
			System.out.println("successfully stole coroutine " + coroutine);
		if (coroutine instanceof Coroutine) {
			Coroutine coro = (Coroutine) coroutine;
			moveCoroutine(coro, scheduledCoroutines);
		}
		coroutine.threadSupport = this;
		coroutine.context = this;
		if (last != null) {
			last.context = other;
		}
		return true;
	}

	CoroutineFrame[] serializeCoroutine(CoroutineBase coroutine) {
		if (coroutine.isFinished()) {
			throw new IllegalArgumentException("Cannot serialize a coroutine that is finished");
		}
		CoroutineSupport other = coroutine.threadSupport;
		if (coroutine == other.threadCoroutine) {
			throw new IllegalArgumentException("Cannot serialize a threads initial coroutine");
		}

		if (!unsafe.compareAndSwapObject(coroutine, coroContextOffset, other, null)) {
			if (TRACE)
				System.out.println("unable to lock coroutine for serializing");
			return null;
		}

		CoroutineFrame[] frames = serializeCoroutineData(coroutine.data);

		coroutine.context = other;

		return frames;
	}

	void replaceCoroutine(CoroutineBase coroutine, CoroutineFrame[] frames) {
		if (coroutine == threadCoroutine) {
			throw new IllegalArgumentException("Cannot replace a threads initial coroutine");
		}

		if (!unsafe.compareAndSwapObject(coroutine, coroContextOffset, this, LOCKED)) {
			coroutineLockFailed(coroutine);
			return;
		}

		if (coroutine instanceof Coroutine) {
			CoroutineFrame[] newFrames = new CoroutineFrame[frames.length + 1];
			System.arraycopy(frames, 0, newFrames, 0, frames.length);
			Method finish = null;
			try {
				String name = frames[frames.length - 1].method.getName();
				if (name.equals("yield")) {
				finish = CoroutineSupport.class.getDeclaredMethod("finishCoroutineSwitch");
					newFrames[frames.length] = new CoroutineFrame(finish, 0, 1, 0, new long[] { 0 }, new Object[] { this });
				} else {
					finish = CoroutineSupport.class.getDeclaredMethod("finishCoroutineSwitch2", new Class[] { Object.class });
					newFrames[frames.length] = new CoroutineFrame(finish, 0, 2, 0, new long[] { 0, 0 }, new Object[] { this, null });
				}
			} catch (Exception e) {
				e.printStackTrace();
			}

			replaceCoroutineData(coroutine.data, newFrames);
		} else {
			replaceCoroutineData(coroutine.data, frames);
		}

		coroutine.context = this;
	}

	public boolean isCurrent(CoroutineBase coroutine) {
		return coroutine == currentCoroutine;
	}

	public CoroutineBase getCurrent() {
		return currentCoroutine;
	}

	private static native void registerNatives();

	private static native long getThreadCoroutine();

	private static native long createCoroutineStack(long stack_size);

	private static native void freeCoroutineStack(long stack);

	private static native long createCoroutine(CoroutineBase coroutine, long stack);

	private static native void replaceCoroutineData(long data, CoroutineFrame[] frames);

	private static native void freeCoroutine(long coroutine);

	private static native void prepareSwitch(CoroutineBase target);

	private static native void switchTo(CoroutineBase current, CoroutineBase target);

	private static native void switchToAndTerminate(CoroutineBase current, CoroutineBase target);

	private static native void switchToAndExit(CoroutineBase current, CoroutineBase target);

	private static native boolean isDisposable(long coroutine);

	private static native CoroutineBase cleanupCoroutine();

	private static native boolean stealCoroutineData(long data, long stack, Thread thread, CoroutineSupport coroutineSupport);

	private static native CoroutineFrame[] serializeCoroutineData(long data);

}
